#!/usr/bin/perl -w

use strict;

use constant PI          => 4*atan2(1,1);
use constant EARTH_RAD_M => 6371000;

sub printOctagon;

my $EARTH_RAD_M = 6371000;

unless ( ( $#ARGV == 0 && -e $ARGV[0] ) ||
         ( $#ARGV == 1 && -e $ARGV[0] && -e $ARGV[1] ) ) {
    die "Usage: cla2kml.pl cla_debug_log [area_monitors]\n"
}

open(FH, $ARGV[0]) or die "Cannot open $ARGV[0]\n";

my ($prevLat, $prevLon) = (100,0);

my @data;
my $i = 0;

while ( <FH> ) {
    if ( m/lat=(\d+\.\d+), lon=(\d+\.\d+), alt=(\d+\.\d+), hacc=(\d+\.\d+)/ ) {
#	next if ( $prevLat == $1 && $prevLon == $2 );
	$data[$i]{"lat"} = $1;
	$data[$i]{"lon"} = $2;
	$data[$i]{"alt"} = $3;
	$data[$i]{"hacc"} = $4;
	$data[$i]{"type"} = 1; #GPS
    } elsif ( m/lat=(\d+\.\d+), lon=(\d+\.\d+), hacc=(\d+\.\d+)/ ) {
#	next if ( $prevLat == $1 && $prevLon == $2 );
	$data[$i]{"lat"} = $1;
	$data[$i]{"lon"} = $2;
	$data[$i]{"alt"} = -1000;
	$data[$i]{"hacc"} = $3;
        $data[$i]{"type"} = ( m/FromMeasurements/ || m/wlanScanReady/ ) ? 2 : 3;
    } else {
	next;
    }

    if ( m/^(\d+)\.(\d+)\.(\d+) (\d+):(\d+):(\d+)\.\d+/ ) {
	$data[$i]{"time"} = sprintf("20%02d-%02d-%02dT%02d:%02d:%02d+02:00", 
				    $3, $2, $1, $4, $5, $6);
    }

    $prevLat = $data[$i]{"lat"};
    $prevLon = $data[$i]{"lon"};
    $i++;
}

close(FH);

my @areamonitor;
if ( $#ARGV == 1 ) {
    open(FH, $ARGV[1]) or die "Cannot open $ARGV[1]\n";
    while ( <FH> ) {
        if ( m/^(-?\d+(\.\d+)?) (-?\d+(\.\d+)?) (\d+(\.\d+)?)$/ ) {
            push(@areamonitor, {lat => $1, lon => $3, rad => $5});
        } 
    }
    close(FH);
}

print << "EOF";
<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<Document>
EOF

# PATH STYLE
print << "EOF";
<Style id="pathStyle">
<LineStyle>
<color>7f00ffff</color>
<width>4</width>
</LineStyle>
<PolyStyle>
<color>7f00ff00</color>
</PolyStyle>
</Style>
EOF

# PATH
print << "EOF";
<Folder>
<name>Path</name>
<Placemark>
<styleUrl>#pathStyle</styleUrl>
<LineString>
<tessellate>1</tessellate>
<altitudeMode>clampToGround</altitudeMode>
<coordinates>
EOF

print $_->{"lon"}.",".$_->{"lat"}."\n" for (@data);

print << "EOF";
</coordinates>
</LineString>
</Placemark>
</Folder>
EOF


if ( @areamonitor ) {
# AREA MONITOR STYLE
print << "EOF";
<Style id="areaMonitorStyle">
<PolyStyle>
<color>b0ff4444</color>
<fill>1</fill>
<outline>0</outline>
</PolyStyle>
</Style>
<Folder>
<name>AREA MONITORS</name>
EOF

for ( @areamonitor ) {    
    print 
	"<Placemark>\n",
	"<styleUrl>#areaMonitorStyle</styleUrl>\n",
	"<Polygon>",
        "<altitudeMode>clampToGround</altitudeMode>",
        "<tessellate>1</tessellate>",
        "<outerBoundaryIs>",
        "<LinearRing>",
	"<coordinates>";

    printPolygon($_->{"lat"}, $_->{"lon"}, $_->{"rad"}, 8);

    print
        "</coordinates>",
        "</LinearRing>",
        "</outerBoundaryIs>",
	"</Polygon>\n",
	"</Placemark>\n";


}
print "</Folder>\n";
}


# GPS STYLE AND POINTS
print << "EOF";
<Folder>
<name>GPS</name>
<Style id="gpsPointStyle">
<IconStyle>
<color>ff00ff00</color>
<scale>0.8</scale>
<Icon>
<href>http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png</href>
</Icon>
</IconStyle>
<PolyStyle>
<color>6600ff00</color>
<fill>1</fill>
<outline>0</outline>
</PolyStyle>
<LineStyle>
<color>ff00ff00</color>
<width>4</width>
</LineStyle>
</Style>
EOF

for my $i ( grep { $data[$_]{"type"} == 1 } (0..$#data) ) {
    print 
	"<Placemark>\n",
        "<description>HAcc=".$data[$i]{"hacc"}."m</description>\n",
	"<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
	"<styleUrl>#gpsPointStyle</styleUrl>\n",
	"<Point>",
	"<coordinates>".$data[$i]{"lon"}.",".$data[$i]{"lat"}."</coordinates>\n",
	"</Point>\n",
	"</Placemark>\n";
}

print "</Folder>\n";


# CELL STYLE
print << "EOF";
<Style id="cellPointStyle">
<IconStyle>
<color>ffff00ff</color>
<scale>0.8</scale>
<Icon>
<href>http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png</href>
</Icon>
</IconStyle>
<PolyStyle>
<color>80ff00ff</color>
<fill>1</fill>
<outline>0</outline>
</PolyStyle>
<LineStyle>
<color>ffff00ff</color>
<width>4</width>
</LineStyle>
</Style>
<Style id="badcellPointStyle">
<IconStyle>
<color>ff0000ff</color>
<scale>0.8</scale>
<Icon>
<href>http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png</href>
</Icon>
</IconStyle>
</Style>
EOF

print  "<Folder>\n<name>CELL</name>\n";

for my $i ( grep { $data[$_]{"type"} == 3 } (0..$#data) ) {
    if ( $data[$i]{"hacc"} < 1000 ) {
        print 
            "<Placemark>\n",
            "<description>HAcc=".$data[$i]{"hacc"}."m</description>\n",
            "<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
            "<styleUrl>#cellPointStyle</styleUrl>\n",
            "<Point>",
            "<coordinates>".$data[$i]{"lon"}.",".$data[$i]{"lat"}."</coordinates>\n",
            "</Point>\n",
            "</Placemark>\n";
    } else {
        print 
            "<Placemark>\n",
            "<description>HAcc=".$data[$i]{"hacc"}."m</description>\n",
            "<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
            "<styleUrl>#badcellPointStyle</styleUrl>\n",
            "<Point>",
            "<coordinates>".$data[$i]{"lon"}.",".$data[$i]{"lat"}."</coordinates>\n",
            "</Point>\n",
            "</Placemark>\n";
    }
}

print "</Folder>\n";
print "<Folder>\n<name>CELL HALO</name>\n";

for my $i ( grep { $data[$_]{"type"} == 3 } (0..$#data) ) {
    if ( $data[$i]{"hacc"} < 1000 ) {
        print 
            "<Placemark>\n",
            "<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
            "<styleUrl>#cellPointStyle</styleUrl>\n",
            "<Polygon>",
            "<altitudeMode>clampToGround</altitudeMode>",
            "<tessellate>1</tessellate>",
            "<outerBoundaryIs>",
            "<LinearRing>",
            "<coordinates>";
        printPolygon($data[$i]{"lat"}, $data[$i]{"lon"}, $data[$i]{"hacc"}, 5);
        print
            "</coordinates>",
            "</LinearRing>",
            "</outerBoundaryIs>",
            "</Polygon>\n",
            "</Placemark>\n";
    }
}

print "</Folder>\n";


# WLAN STYLE AND POINTS
print << "EOF";
<Style id="wlanPointStyle">
<IconStyle>
<color>ffffff00</color>
<scale>0.8</scale>
<Icon>
<href>http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png</href>
</Icon>
</IconStyle>
<PolyStyle>
<color>a0ffff00</color>
<fill>1</fill>
<outline>0</outline>
</PolyStyle>
<LineStyle>
<color>ffffff00</color>
<width>4</width>
</LineStyle>
</Style>
EOF

print "<Folder>\n<name>WLAN</name>\n";

for my $i ( grep { $data[$_]{"type"} == 2 } (0..$#data) ) {
    print 
        "<Placemark>\n",
        "<description>HAcc=".$data[$i]{"hacc"}."m</description>\n",
        "<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
        "<styleUrl>#wlanPointStyle</styleUrl>\n",
        "<Point>",
        "<coordinates>".$data[$i]{"lon"}.",".$data[$i]{"lat"}."</coordinates>\n",
        "</Point>\n",
        "</Placemark>\n";
}

print "</Folder>\n";
print "<Folder>\n<name>WLAN HALO</name>\n";

for my $i ( grep { $data[$_]{"type"} == 2 } (0..$#data) ) {
    print 
	"<Placemark>\n",
	"<TimeStamp><when>".$data[$i]{"time"}."</when></TimeStamp>\n",
	"<styleUrl>#wlanPointStyle</styleUrl>\n",
	"<Polygon>",
        "<altitudeMode>clampToGround</altitudeMode>",
        "<tessellate>1</tessellate>",
        "<outerBoundaryIs>",
        "<LinearRing>",
	"<coordinates>";
    printPolygon($data[$i]{"lat"}, $data[$i]{"lon"}, $data[$i]{"hacc"}, 6);
    print
        "</coordinates>",
        "</LinearRing>",
        "</outerBoundaryIs>",
	"</Polygon>\n",
	"</Placemark>\n";
}

print "</Folder>\n";


print << "EOF";
</Document>
</kml>
EOF

exit(0);

sub printPolygon
{
    my ($lat, $lon, $rad, $n) = @_;
    my $lat_rad = ($rad/EARTH_RAD_M)*(180.0/PI);
    my $lon_rad = ($rad/(EARTH_RAD_M*cos(PI*$lat/180.0)))*(180.0/PI);
    for ( my $i=0; $i<$n; $i++ ) {
        printf("%.6f,%.6f,1 ",
               $lon+$lon_rad*sin(2*$i*PI/$n),
               $lat+$lat_rad*cos(2*$i*PI/$n));
    }
    printf("%.6f,%.6f,1", $lon, $lat+$lat_rad);
}
